/* *****************************************************************************
 * pather.c 1.12 : world navigation routines.                                  * 
 * This code was originally written in 1995 as a standalone tool.  Ported into *
 * LO in 2011 under The Last Outpost Project license.                          *
 * version 1.12 is released by the author rahjiii@jeffrika.com under LGPL      *
 ***************************************************************************** */

/*char RCSid_pather_c[] = "$Header: $"; */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#include "structs.h"
#include "db.h"
#include "utils.h"

/* NOT IN DIKU GAMMA */
/* #include "constants.h" */
/* #include "debug.h" */

#include "pather.h"

/* A BUNCH OF DEFINES AND DECLARATIONS TO GET THIS TO COMPILE UNDER DIKU GAMMA */
/* call pather_init() sometime after the world has been loaded up, then you are
 * clear to call whichway() or run_directions() all you want. */
extern int top_of_world;
extern struct room_data *world;
#define EX_DYNAMIC	0x80
#define EX_PORTAL	0x400
extern int movement_loss[];
const char *dirshead[] =
{
    "to the North",
    "to the East",
    "to the South",
    "to the West",
    "up from here",
    "down from here",
    "\n"
};
#define lo_debug(facility,args...) {}
#define exit_order(order,index) (index);
/* END GOOFY DIKU HACKS BEGIN GOOFY LO HACKS*/


/* local defines */
#define ENQUEUE 0
#define DEQUEUE 1
#define INIT 2
#define MAX_DIST 32767

/* Some tunable parameters for data structure caching.  The apsp() routine can
 * iterate over the whole world, so it pays to try to avoid calling it
 * frequently.  The original whichway cache implementation didnt have a timer
 * on the cache entries, but that was before boats and portals were added to
 * the game.  They can change how the world is connected up, so it may be good
 * to not cache results for too long. WWAY_CACHE_DYNSEC is the cache timeout
 * when a path mode that can cross through a dynamic door is used.  Otherwise,
 * WWAY_CACHE_SECONDS is used, and it can be very long. */
#define WWAY_CACHE_SIZE 2039
#define WWAY_CACHE_DYNSEC 60
#define WWAY_CACHE_SECONDS 3600

/* external vars */

/* some globals.*/
struct whichway_cache_struct *whichway_cache = NULL;

/* local forward function declarations */
int whichway_fetch(int from, int to, int mode, int *whichway);
int whichway_store(int from, int to, int mode,int whichway);
int whichway_cache_key(int from, int to, int mode);

/* code follows... */

/* return an index into the hash table. */
int whichway_cache_key(int from, int to, int mode) {
	return( ((mode<<16) + (to<<12) + (from)) % WWAY_CACHE_SIZE );
}

/* Initialize the data structures.  This is mostly just to init (or re-init)
 * the cache. Call this once at boot time before using any of the other path
 * calls.  */
void pather_init(void) {

	int i;
	
	if(whichway_cache) {
		free(whichway_cache);
	} 

	CREATE(whichway_cache, struct whichway_cache_struct, WWAY_CACHE_SIZE);
	for(i=0;i<WWAY_CACHE_SIZE;i++) {
		whichway_cache[i].from = -1;
		whichway_cache[i].to = -1;
		whichway_cache[i].mode = -1;
		whichway_cache[i].expires = 0;
		whichway_cache[i].whichway = -1;
	}

	lo_debug(DEBUG_PATHER,"whichway cache is %d bytes",
		(sizeof(struct whichway_cache_struct) * WWAY_CACHE_SIZE)
	);
}

/* A simple queue for storing rooms that need to be probed. I'm sure I copied
 * this implementation from somewhere... */
int apsp_queue(int command, int room, int dist) {

	struct node {
		int room;
		int dist;
		struct node *next;
	}; 

	static struct node *head = NULL;
	struct node *aptr;
	struct node *bptr;
	int retval;
	
	switch (command) {
		default:
		case INIT:
			while(head != NULL) {
				aptr = head;
				head = head->next;
				free(aptr);
			}
			return(1);

		case DEQUEUE:
			if(head == NULL) return(-1);
			retval = head->room;
			aptr = head;
			head = head->next;
			free(aptr);
			return(retval);

		case ENQUEUE:

			aptr = (struct node*) malloc(sizeof(struct node));
			aptr->next = NULL;
			aptr->room = room;
			aptr->dist = dist;
			
			if(head == NULL){
				head=aptr;
				return(room);
			}

			if (head->dist >= aptr->dist) {
				aptr->next = head;
				head = aptr;
				return(room);
			}

			bptr = head;
			while (bptr->next) {
				if(bptr->next->dist >= aptr->dist) {
					aptr->next = bptr->next;
					bptr->next = aptr;
					return(room);
				}
				bptr = bptr->next;
			}
			bptr->next = aptr;
			return(room);

	}
}

/* All Points Shortest Path.  Call apsp() with the real room where you ARE.
 * This gives you a structure that can give you the shortest path to everywhere
 * in the world.  To get to somewhere, find the destination room in the
 * allpoints structure, and if the room is marked as known, follow the chain of
 * camefroms back to where you ARE. Free it when you are done with it.*/
struct allpoints *apsp(int seed,int mode) {

	int room;
	int adj;
	int i,exit;
	int mybet;
	struct allpoints *ap;
	struct room_direction_data *door;
	int roomcount = 0;
	int doorcount = 0;

	CREATE(ap,struct allpoints,top_of_world+1);

	lo_debug(DEBUG_PATHER,"seed room %d mode %d",virtual_room(seed),mode);

	for(i=0;i<=top_of_world;i++){
		ap[i].known = -1;
		ap[i].dist = MAX_DIST;
		ap[i].camefrom = -1;
		ap[i].bygoing = -1;
	}

	apsp_queue(INIT,0,0);
	ap[seed].dist = 0;
	apsp_queue(ENQUEUE,seed,0);

	while((room=apsp_queue(DEQUEUE,0,0)) != -1) {
		if(ap[room].known != 1) {
			ap[room].known=1;
			roomcount++;
			for(i=0;i<6;i++) {
				exit = exit_order(room,i);
				if( (door = world[room].dir_option[exit]) ) {
					adj=door->to_room;
					if( (adj > top_of_world) || 
						(adj < 0) ||
						((mode & PATH_NOSWIM) && (world[adj].sector_type == SECT_WATER_NOSWIM)) ||
						((mode & PATH_STAYZONE) && (world[adj].zone != world[room].zone)) ||
						((mode & PATH_NOCLOSED) && (IS_SET(door->exit_info,EX_CLOSED))) ||
						((mode & PATH_NOLOCKED) && (IS_SET(door->exit_info,EX_LOCKED))) ||
						(!(mode & PATH_OKPORTAL) && (IS_SET(door->exit_info,EX_PORTAL))) ||
						(!(mode & PATH_OKDYNAMIC) && (IS_SET(door->exit_info,EX_DYNAMIC)))
					) {
						/* do nothing. */
						continue;
						lo_debug(DEBUG_PATHER,"from %d exit %d, not probing into room %d",
							virtual_room(room),exit,virtual_room(adj)
						);
					} else if((adj >= 0) && (ap[adj].known != 1)) {
						mybet = ap[room].dist;
						if(mode & PATH_NOMOVELOSS) {
							/* add one to the length of the path. */
							mybet +=1;
						} else {
							/* add what it would normally cost in mv to go between the rooms. */
							mybet += (
								(movement_loss[world[room].sector_type]
									+ movement_loss[world[adj].sector_type] 
								) /2
							);
						}
						/* preturb the distance to make the paths fuzzier: mybet+=i; */
						if(mybet < ap[adj].dist) {
							ap[adj].dist = mybet;
							ap[adj].camefrom = room;
							ap[adj].bygoing = exit;
							apsp_queue(ENQUEUE,adj,mybet);
							doorcount++;
						}
					}
				}
			} /*end of forloop*/
		}
	}
	lo_debug(DEBUG_PATHER,"roomcount %d doorcount %d",roomcount,doorcount);
	return(ap);
}

/* looks in the whichway cache for a valid entry.  If found, stores it in
 * *whichway and returns 1.  Returns 0 otherwise. */
int whichway_fetch(int from, int to, int mode, int *whichway) {
	int key;
	int ret;
	static unsigned int calls = 0;
	static unsigned int hits = 0;

	calls++;
	key = whichway_cache_key(from,to,mode);

	if( (whichway_cache[key].from == from) &&
		(whichway_cache[key].to == to) &&
		(whichway_cache[key].mode == mode) &&
		(whichway_cache[key].expires > time(NULL))
	) {
		/* you can safely call with whichway==NULL if you don't want the return value. */
		if(whichway != NULL) {
			*whichway = whichway_cache[key].whichway;
		}
		hits++;
		ret=1;
	} else {
		ret=0;
	}
	lo_debug(DEBUG_PATHER,"%d,%d,%d cache %s (%3.2f%%).",
		virtual_room(from),
		virtual_room(to),
		mode,
		ret?"hit":"miss",
		100.0 * hits/calls
	);
	return(ret);
}

/* puts an entry into the whichway cache. */
int whichway_store(int from, int to, int mode,int whichway) {
	int key;

	key = whichway_cache_key(from,to,mode);
	whichway_cache[key].from = from;
	whichway_cache[key].to = to;
	whichway_cache[key].mode = mode;
	whichway_cache[key].whichway = whichway;

	/* If the path mode is ok with traversing dynamic doors, use a shorter
	 * cache expiry time.  Its not really a big deal if a cached path contains
	 * a door that has gone away. A path follower might end up walking to where
	 * the door used to be before being led on a different path to the
	 * destination. *shrug* */
	if( (mode & PATH_OKDYNAMIC) ||
		(mode & PATH_OKPORTAL)
	) {
		whichway_cache[key].expires = time(NULL) + WWAY_CACHE_DYNSEC;
	} else {
		whichway_cache[key].expires = time(NULL) + WWAY_CACHE_SECONDS;
	}

	lo_debug(DEBUG_PATHER,"%d,%d,%d = %d",
		virtual_room(from),
		virtual_room(to),
		mode,
		whichway_cache[key].whichway
	);
	return(key);

}

/* whichway() returns the best exit direction to use, given real room numbers
 * from and to.  returns -1 if you can't get there from here, or are already
 * there. Mode is a bitmask of PATH_traversal options.  If you are following
 * whichway()'s adivce, repeated calls to it are pretty fast, because it is
 * trying to cache the path for you.  */
int whichway(int from, int to, int mode) {

	struct allpoints *ap;
	int bygoing;
	int hop;

	if( (from < 0) ) return(-1);
	if( (to < 0) ) return(-1);
	if(to == from) return(-1);

	if (whichway_fetch(from, to, mode, &bygoing)) {
		if( (bygoing == -1)  || 
			(world[from].dir_option[bygoing] != NULL)
		) {
			return(bygoing);
		}
	}

	ap = apsp(from,mode);

	bygoing = -1;
	if(ap[to].known != 1) {
		lo_debug(DEBUG_PATHER,"whichway: No path from #%d to #%d.",
			virtual_room(from),
			virtual_room(to)
		);
		whichway_store(from,to,mode,-1);
	} else {
		hop = to;
		while(hop != from) {
			bygoing = ap[hop].bygoing;
			hop = ap[hop].camefrom;
			whichway_store(hop,to,mode,bygoing);
		}
	}
	free(ap);
	return(bygoing);

}

/* Given a string of direction characters like "nnnnnneeeww", try to compress
 * it down into a more compact format like "6n3e2w", which is easier for a
 * player to read and use. Remember to free what this thing gives you. */
char *compressed_directions(char *path) {

	char *ret, *r, *s;
	char d;
	int run;

	ret = strdup(path);

	/* and compress it. */
	r = path;
	s = ret;
	run = 0;
	while( (d = *r++) ) {
		run++;
		if(*r != d) {
			if(run>1) {
				s += sprintf(s,"%d%c",run,d);
			} else {
				s += sprintf(s,"%c",d);
			}
			run=0;
		}
	}

	lo_debug(DEBUG_PATHER,"compressed: %s",ret);

	return(ret);
} 



/* run_directions() returns an uncompressed string of dirction chars (i.e. no
 * numbers for repeated directions) that'll take you from and to. Caller is
 * responsible for freeing this string when done with it.  
 */
char *run_directions(int from, int to,int mode) {

	int bygoing;
	char buf[MAX_STRING_LENGTH];
	char *s,*eob;
	char dirs[] = "neswud"; /* make this a constants.h global sometime... */
	char *ret;

	if( (from < 0) ) return(strdup(""));
	if( (to < 0) ) return(strdup(""));

	lo_debug(DEBUG_PATHER,"finding path from #%d to #%d.",
		virtual_room(from),
		virtual_room(to)
	);

	if( (whichway(from,to,mode) == -1) ) {
		lo_debug(DEBUG_PATHER,"No path from #%d to #%d.",
			virtual_room(from),
			virtual_room(to)
		);
		return(strdup(""));
	}

	if(to == from) {
		return(strdup(""));
	}

	s = buf;
	eob = buf + sizeof(buf)-2;

	while( (from != to) && (s!=eob)) {
		bygoing = whichway(from,to,mode);
		if(bygoing == -1) {
			lo_debug(DEBUG_PATHER,"A cached path has disappeared due to dynamic door change.");
			break;
		} else {
			*s++ = dirs[bygoing];
			from = world[from].dir_option[bygoing]->to_room;
		}
	}
	*s = '\0';

	ret=strdup(buf);
	lo_debug(DEBUG_PATHER,"uncompressed: %s",ret);

	return(ret);

}

/* put a text description of where to go next into buf, given real room numbers from and to. */
char *whichheading(int from, int to, int mode, char *buf) {

	int dir;

	if(from==to) {
		sprintf(buf,"right here");
	} else if((dir=whichway(from,to,mode)) == -1) {
		sprintf(buf,"elsewhere");
	} else {
		sprintf(buf,"%s",
			/* (IS_SET(world[from].room_flags, NAUTICAL))?nautdirshead[dir]:dirshead[dir] */
			dirshead[dir]
		);
	}

	return(buf);

}

/* Utility command that maps a cardinal direction char to the associated do_XXX
 * style cmd number.  So in a mob program you can do something like:  
	do_move(me,"",cmd_from_dirchar('n'));
*/
int cmd_from_dirchar(char d) {
	switch(d) {
	case 'n': return(1);
	case 'e': return(2);
	case 's': return(3);
	case 'w': return(4);
	case 'u': return(5);
	case 'd': return(6);
	default: return(0);
	}
}
